
// version, APPLE, SHADOW_SHADER, SHADOW_MAP_BIAS and MAX_BONE_NUM are defined by application

#ifdef APPLE
	#define IN attribute
	#define OUT varying
	#define INT float
	#define IVEC4 vec4
#else
	#define IN in
	#define OUT out
	#define INT int
	#define IVEC4 ivec4
	#define ATTR_INT
#endif

IN vec3 vPosition;
IN vec3 vNormal;
IN vec2 vUV;
IN float boneWeightType;
IN IVEC4 iBone;
IN vec4 fBoneWeight;
IN vec3 sdefC;
IN vec3 sdefR0;
IN vec3 sdefR1;
IN vec3 vFace;
#ifdef APPLE
	bindable uniform vec4 boneMatrixArray[MAX_BONE_NUM * 3];
#else
	layout(std140) uniform BoneMatrixArray {
		vec4 boneMatrixArray[MAX_BONE_NUM * 3];
	};
#endif

#ifdef APPLE
	OUT vec2 vTexCoord;
	OUT vec3 normal;
#else
	OUT vData {
		vec2 vTexCoord;
		vec3 normal;
	};
#endif

mat4 unpackMatrix(INT id) {
	#ifdef ATTR_INT
		int index = id;
	#else
		int index = int(id + 0.1);
	#endif
	return transpose(mat4(boneMatrixArray[index * 3], 
							boneMatrixArray[index * 3 + 1], 
							boneMatrixArray[index * 3 + 2], 
							vec4(0.0, 0.0, 0.0, 1.0)));
}

vec4 matrixToQuaternion(mat4 m4) {
	float m00=m4[0][0],m01=m4[0][1],m02=m4[0][2]
	     ,m10=m4[1][0],m11=m4[1][1],m12=m4[1][2]
		 ,m20=m4[2][0],m21=m4[2][1],m22=m4[2][2];
	float t = m00 + m11 + m22;
	vec4 result=vec4(0,0,0,1);
	float x,y,z,w;
	if (t >= 0) 
	{ 
		float s = sqrt(t + 1); 
		w = 0.5f * s;
		s = 0.5f / s;                 
		x = (m21 - m12) * s;
		y = (m02 - m20) * s;
		z = (m10 - m01) * s;
	} 
	else if ((m00 > m11) && (m00 > m22)) 
	{
		float s = sqrt(1.0f + m00 - m11 - m22); 
		x = s * 0.5f; 
		s = 0.5f / s;
		y = (m10 + m01) * s;
		z = (m02 + m20) * s;
		w = (m21 - m12) * s;
	} 
	else if (m11 > m22) 
	{
		float s = sqrt(1.0f + m11 - m00 - m22); 
		y = s * 0.5f; 
		s = 0.5f / s;
		x = (m10 + m01) * s;
		z = (m21 + m12) * s;
		w = (m02 - m20) * s;
	} 
	else 
	{
		float s = sqrt(1.0f + m22 - m00 - m11); 
		z = s * 0.5f; 
		s = 0.5f / s;
		x = (m02 + m20) * s;
		y = (m21 + m12) * s;
		w = (m10 - m01) * s;
	}
	result.x=x;
	result.y=y;
	result.z=z;
	result.w=w;
	return result;
}

mat4 quaternionToMatrix(vec4 quat) {
	mat4 result=mat4(1.0);
	float x=quat.x,y=quat.y,z=quat.z,w=quat.w;
	float norm = w * w + x * x + y * y + z * z;
	float s = (norm == 1) ? 2 : (norm > 0) ? 2 / norm : 0;
	float xs = x * s;
	float ys = y * s;
	float zs = z * s;
	float xx = x * xs;
	float xy = x * ys;
	float xz = x * zs;
	float xw = w * xs;
	float yy = y * ys;
	float yz = y * zs;
	float yw = w * ys;
	float zz = z * zs;
	float zw = w * zs;
	result[0][0] = 1 - (yy + zz);
	result[0][1] = (xy - zw);
	result[0][2] = (xz + yw);
	result[1][0] = (xy + zw);
	result[1][1] = 1 - (xx + zz);
	result[1][2] = (yz - xw);
	result[2][0] = (xz - yw);
	result[2][1] = (yz + xw);
	result[2][2] = 1 - (xx + yy);
	return result;
}

vec4 slerp(vec4 q1,vec4 q2,float t) {
	if (q1.x == q2.x && q1.y == q2.y && q1.z == q2.z && q1.w == q2.w) 
	{
		return q1;
	}

	float result = (q1.x * q2.x) + (q1.y * q2.y) + (q1.z * q2.z)+ (q1.w * q2.w);
	if (result < 0.0f) 
	{
		q2.x = -q2.x;
		q2.y = -q2.y;
		q2.z = -q2.z;
		q2.w = -q2.w;
		result = -result;
	}
	float scale0 = 1 - t;
	float scale1 = t;
	if ((1 - result) > 0.1f) 
	{
		float theta = acos(result);
		float invSinTheta = 1 / sin(theta);
		scale0 = sin((1 - t) * theta) * invSinTheta;
		scale1 = sin((t * theta)) * invSinTheta;
	}
	vec4 resultQuat=vec4(0,0,0,1);
	resultQuat.x = (scale0 * q1.x) + (scale1 * q2.x);
	resultQuat.y = (scale0 * q1.y) + (scale1 * q2.y);
	resultQuat.z = (scale0 * q1.z) + (scale1 * q2.z);
	resultQuat.w = (scale0 * q1.w) + (scale1 * q2.w);
	return resultQuat;
}

#define SUM_POSITION_WITH_WEIGHT(INDEX, MPOS, EPOS, MNORMAL, ENORMAL, TMPMAT) \
		TMPMAT = unpackMatrix(iBone[INDEX]); \
		EPOS += TMPMAT * MPOS * fBoneWeight[INDEX]; \
		ENORMAL += TMPMAT * MNORMAL * fBoneWeight[INDEX];

vec4 getPosition() {
	vec4 mPos = vec4(vPosition+vFace,1.0);
	vec4 mNormal = vec4(vNormal, 0.0);
	vec4 ePos = vec4(0.0);
	vec4 eNormal = vec4(0.0);
	if(boneWeightType==0)//BDEF
	{
		mat4 tempMatrix;
		SUM_POSITION_WITH_WEIGHT(0, mPos, ePos, mNormal, eNormal, tempMatrix);
	}
	else if(boneWeightType==1)//BDEF2
	{
		mat4 tempMatrix;
		SUM_POSITION_WITH_WEIGHT(0, mPos, ePos, mNormal, eNormal, tempMatrix);
		SUM_POSITION_WITH_WEIGHT(1, mPos, ePos, mNormal, eNormal, tempMatrix);
	}
	else if(boneWeightType==2)//BDEF4
	{
		mat4 tempMatrix;
		SUM_POSITION_WITH_WEIGHT(0, mPos, ePos, mNormal, eNormal, tempMatrix);
		SUM_POSITION_WITH_WEIGHT(1, mPos, ePos, mNormal, eNormal, tempMatrix);
		SUM_POSITION_WITH_WEIGHT(2, mPos, ePos, mNormal, eNormal, tempMatrix);
		SUM_POSITION_WITH_WEIGHT(3, mPos, ePos, mNormal, eNormal, tempMatrix);
	}
	else if(boneWeightType==3)//SDEF
	{
		mat4 boneMatrix0 = unpackMatrix(iBone[0]), boneMatrix1 = unpackMatrix(iBone[1]);
		vec3 sdefRW0 = sdefR0;
		vec3 sdefRW1 = sdefR1;
		vec3 vector = vec3(boneMatrix0*vec4(sdefRW0,0));
		vec3 vector2 = vec3(boneMatrix1*vec4(sdefRW1,0));
		vec3 vector3 = vec3(boneMatrix0*vec4(sdefC,1.0));
		vec3 vector4 = vec3(boneMatrix1*vec4(sdefC,1.0));
		vec3 vector5 = (fBoneWeight[0] * ((vector3 - sdefC) + (0.5 * vector))
		             + (fBoneWeight[1]) * ((vector4 - sdefC) + (0.5 * vector2))) 
					 + sdefC;
		mat4 matrix = quaternionToMatrix( slerp(matrixToQuaternion(boneMatrix0), matrixToQuaternion(boneMatrix1), fBoneWeight[1]));
		eNormal = matrix * mNormal;
		ePos = vec4((matrix * vec4(mPos.xyz - sdefC, 0.0)).xyz + vector5, 1.0);
	}
	vTexCoord = vUV;
	normal = normalize(gl_NormalMatrix * eNormal.xyz);
	return gl_ModelViewMatrix * ePos;
}

void main() {
#ifdef SHADOW_SHADER
	gl_Position = gl_ProjectionMatrix * getPosition();
	float dist = sqrt(gl_Position.x * gl_Position.x + gl_Position.y * gl_Position.y);
	float distortFactor = (1.0f - SHADOW_MAP_BIAS) + dist * SHADOW_MAP_BIAS;
	#ifdef SEUS110
		gl_Position.xy *= 0.95f / distortFactor;
		gl_Position.z = mix(gl_Position.z, 0.5, 0.8);
		//gl_FrontColor = gl_Color;
	#else
		gl_Position.xy *= 1.0f / distortFactor;
	#endif
#else
	gl_Position = getPosition();
#endif
}